
/* Define this in the MAKEFILE, not here */
/* #define DEBUG */ 

#include "sunddk.h" /* #defines the crucial SUNSOFT_BUYBACK magic token */

/* This stuff comes from SUNWxwddk */
#define NEED_EVENTS
#include    "os.h"
#include    "sun.h"
#include    "suninput.h"
#include    "input.h"
#include    "mi.h"

#include <fcntl.h>

#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <stropts.h>
#include <sys/conf.h>
#include <sys/stat.h>

#include <X11/Sunowconfig.h>


static char *version="1.9 ddxBOLTusbsm.c";

/* extension module for Xsun, to use data from /dev/usbsm0
 *   [See www.bolthole.com/solaris/drivers/]
 *
 * Note that this module is SPECIFICALLY FOR Xsun.
 * You cannot compile it into XFree or vanilla X11R6, without significant
 * additions/wrappers for it, since there is a lack of standardization
 * for Xserver module entry points.
 *
 * Note also that you NEED THE SUN X DDX to compile it.
 * The DDK can be freely downloaded from soldc.sun.com somewheres.
 *
 *
 * Top-level entry point is ddxBOLTusbsmProc()
 * Then usbsmGetEvents() and usbsmEnqueue() do all the real work
 *
 * Code in here was devolved from my original Wacom Graphire tablet driver
 */


/* Size of a complete, raw datapacket from the USB mouse.
 */
#define RAW_EVENT_SIZE 4

/* This holds "driver private" data for the tablet.
 * The server holds on to this for us, and passes it back to us via
 * a DevicePtr->devicePrivate pointer
 */
typedef struct {
	int fd;
	unsigned char lastbutton;
} usbsmPriv;


/*  usbsmGetEvents().
 * Can be considered what the Sun X docs refer to as
 *       DeviceReadProc()
 *
 * Called when the server has received a SIGPOLL for our device.
 *  (Although how it knows it is for OUR device, I'm not sure)
 * Pass back to the server a pointer to an array of firmware events,
 * and update the pointer-to-int to let it know how many.
 * NULL/0 are acceptible values.
 *
 * callAgain is set to true, if we return before reading all the
 *   data that is waiting for us on the device. In that case, the
 *   server will call us again, to finish the processing.
 *
 * In theory, the server will receive a signal for every new mouse event.
 * However, it is unknown if we will see every signal. So we have to
 * be paranoid, I guess.
 * For now, just read one event at a time?
 * XXX later versions, read up to some max number, maybe.
 *
 * Apparently, the Xserver will not call us back until it has used up
 * all the events we have returned from a previous call.
 * So it is acceptible to use static event buffers here.
 *
 * This all seems very wierd. I think that we have a very quick, short
 * call here to more accurately associate timestamps with each event.
 * Then we do the actual conversion batch-style, later.
 *  (No wonder X pointer latency is often horrible)
 */
#define MAX_EV_READ 5
static XI_eventPtr
usbsmGetEvents(DevicePtr X11devptr, int *numEvents, Bool *callAgain)
{
	struct timeval timenow;
	usbsmPriv *privdata=(usbsmPriv*)X11devptr->devicePrivate;
	static XI_event eventbuff[MAX_EV_READ];
	int readstat;
	static CARD8 rawevents[MAX_EV_READ][RAW_EVENT_SIZE];
	int maxbytes=MAX_EV_READ * RAW_EVENT_SIZE;
	int rawcount;

	/* Doing an ioctl(NREAD) doesnt seem to work too well.
	 * So just mimic what sample sunTablet does: read as much
	 * as possible, since we have already set FNDELAY in opendev()
	 */
	readstat=read(privdata->fd,rawevents,maxbytes);
	if(readstat==maxbytes){
		*callAgain=TRUE;
	} else {
		*callAgain=FALSE;
	}
	*numEvents= readstat / RAW_EVENT_SIZE;

	gettimeofday(&timenow,NULL);

	for(rawcount=0 ; rawcount< *numEvents ; rawcount++){
		eventbuff[rawcount].time=timenow;
		eventbuff[rawcount].event=&rawevents[rawcount];
	}

	return (eventbuff);
}

/* This is split out from usbsmEnqueue, to keep that routine clean.
 * We've somehow determined that we need to send a button event, rather than
 * just tweak pointer positoin. So now we need to figure out
 * what TYPE of event: ButtonPress vs ButtonRelease, and
 * which button.
 * We make the "bad" assumption that it is impossible for two buttons
 * to change state at the exact same time.
 *
 * We only update u.u.type, and u.u.detail.
 * Everything else in the event struct is set by the caller
 */
static void calculateButtonEvent(
	unsigned char oldbuttons,
	unsigned char newbuttons,
	struct _xEvent *eventptr)
{
	int flag=0;
	int buttonlist=0;

	/* Note that we FLIP buttons 2 and 3.
	 * /dev/usbsm0 gives us the RAW 4byte packet from a standard HID
	 * compliant mouse. For some reason, the middle button tends to
	 * be encoded as button 3, but UNIX likes it to be button 2
	 */
	if((oldbuttons&0x1) != (newbuttons&0x1)){
		flag=newbuttons & 0x01;
		buttonlist=Button1;
	} else if((oldbuttons&0x2) != (newbuttons&0x2)){
		flag=newbuttons & 0x02;
		buttonlist=Button3; /* FLIP */
	} else if((oldbuttons&0x4) != (newbuttons&0x4)){
		flag=newbuttons & 0x04;
		buttonlist=Button2; /* FLIP */
	} else if((oldbuttons&0x8) != (newbuttons&0x8)){
		flag=newbuttons & 0x08;
		buttonlist=Button4;
	} else if((oldbuttons&0x10) != (newbuttons&0x10)){
		flag=newbuttons & 0x10;
		buttonlist=Button5;
	}

	eventptr->u.u.detail=buttonlist;

	if(flag >0){
#ifdef DEBUG
		ErrorF("ddxBOLTusbsm doing buttonPRESS\n");
#endif
		eventptr->u.u.type=ButtonPress;
	} else {
#ifdef DEBUG
		ErrorF("ddxBOLTusbsm doing buttonRelease for button\n");
#endif
		eventptr->u.u.type=ButtonRelease;
	}
}

/* We are passed ONE 'raw' event by the Xserver.
 * This is the exact same data we ourselves gave to the Xserver
 * from usbsmGetEvents()
 *
 * Our task here is to either do something directly, or convert
 * the 'raw' data, into a Xserver standard event packet.
 * Damn stupid architecture, but thats what I have got to work with.
 *
 * I think the reason this doubleback is done, is a latency issue.
 * For some reason, the server coders decided that it was important to
 * shove the timestamps on the raw data first.
 *
 * For mouse movement, we just tell the server directly,
 * "Hey! Update the core pointer!"
 * For button press/release, we have to actually queue up an event.
 */
static void
usbsmEnqueue(DevicePtr X11devptr, XI_eventPtr rawev_ptr)
{
	usbsmPriv *privdata=(usbsmPriv*)X11devptr->devicePrivate;
	int x,y;
	unsigned long timestamp;
	int rootx,rooty;
	unsigned char newbutton;
	CARD8 *devdata=rawev_ptr->event;
	/* ugliness done to make alignment work out right, or something */
	xEvent eventalloc[3];
	struct _xEvent *neweventptr=&eventalloc[0];

	/*srcdevid=((DeviceIntPtr)X11devptr)->id;*/

	x=(int8_t)devdata[1];
	y=(int8_t)devdata[2];

	timestamp=TVTOMILLI(rawev_ptr->time);
	miPointerDeltaCursor(x,y,timestamp);

	newbutton=devdata[0] & 0x07;

	/* We have to do a little translation on the "raw" data.
	 * the wheel presents itself like a third 'axis', with
	 * a ludicrously small range of movement.
	 * Convert it instead to look like Button4 or Button5 events.
	 * HOWEVER.. only do the hack, if the middle button is NOT pressed.
	 * This avoids the phenomenon of normally pressing in the button
	 * for button2 events, and accidentally rotating the wheel and
	 * getting "extra" buttonclicks
	 *
	 */
	if(newbutton==0){
		switch(devdata[3]){
		    case 0x01:
			newbutton|= 0x08; /* button4 == "wheel up" */
			break;
		    case 0xff:
			newbutton|= 0x10; /* button5 == "wheel down" */
			break;
		}
	}

 	if(newbutton == privdata->lastbutton){
		/* no button state change. We are done here */
		return; 
	}

	/* This should really be called miGETPointerPosition */
	miPointerPosition(&rootx, &rooty);

	neweventptr->u.keyButtonPointer.time=timestamp;

	/* set info in KBTevent */
	calculateButtonEvent(privdata->lastbutton, newbutton,
		                     neweventptr);
	privdata->lastbutton=newbutton;


#ifdef DEBUG
	ErrorF("usbsmEnqueue: x=%d, y=%d: newbutton=%x\n",x,y,newbutton);
	ErrorF("Sending KBTevent of type %d for devid %d\n",
	       KBTevent->type,KBTevent->deviceid);
#endif

	/* Note: The server memcopies the data in KBTevent */
	mieqEnqueue(&eventalloc[0]);

}


/* We are required to have a function like this, to pass in to
 * InitValuatorClassDeviceStruct()
 * However, since we we do not record motion history, we dont have to
 * actually DO anything here.
 */
static int
usbsmMotionProc(DeviceIntPtr ptr, xTimecoord *coords,
                unsigned long start, unsigned long stop,
                ScreenPtr s_ptr)
{
	return 0;
}


/* Open device. Called by ddxBOLTusbsmProc()
 * Set I/O functioning of device.
 * Set (FASYNC | FNDELAY)
 * return 0 on FAIL, 1 on good
 */
static int
opendev(DevicePtr X11devptr, char *devname)
{
	int newfd=-1;

	usbsmPriv *privdata=(usbsmPriv*)X11devptr->devicePrivate;

	if(privdata==NULL){
		Error("BOLTusbsm:opendev: NULL privdata?");
		return (0);
	}

	newfd=open(devname, O_RDONLY);
	if(newfd <0){
		perror("BOLTusbsm");
		ErrorF("BOLTusbsm:opendev: could not open device\n");
		return (0);
	}
	privdata->fd=newfd;

	if (fcntl (privdata->fd, F_SETFL, (FNDELAY|FASYNC)) < 0) {
		perror("BOLTusbsm");
		ErrorF("Can't set up tablet on fd %d\n", privdata->fd);
		return (0);

	}
	return (1);
}

/* I dont know why this is called "AxisChange". I'm just copying
 * the name from sunTablet.
 * It should really be called "CorePointerNotify" or similar
 */
static int
usbsmAxisChange(DevicePtr X11devptr, Bool isCore, unsigned char x, 
		    unsigned char y)
{
	if(isCore){
		ErrorF("ddxBOLTusbsm acknowleging Core status\n");
		mieqUpdateKbdPtr(NULL, X11devptr);
	} else {
		ErrorF("ddxBOLTusbsm no longer Core\n");
	}
	return (Success);
}

/* I think this exists for when an X11WarpPointer() call is done.
 * The X server sets its own internal pointer coordinates, and
 * then tells us to do the same.
 * Except we dont STORE absolute coordinates right now, so ignore
 */
int usbsmSetValuators(DevicePtr X11devptr, int *valuators, 
        int first, int num_valuators)
{
	return (Success);
}

/* This routine is to allow for mouse acceleration and "threshhold" 
 * adjusting.
 * Too bad we dont support that sort of thing right now.
 */
void usbsmPtrCtrlProc(DeviceIntPtr devintptr, PtrCtrl *pointerctl)
{
	/* pointerctl has fields num, den, threshhold */
#ifdef DEBUG
	ErrorF("DEBUG: usbsmPtrCtrlProc called\n");
#endif

	return;
}

/*
 * Register ourselves with X server.
 * Called as part of our DEVICE_INIT case handling.
 * Initialize server private stuff, and bind ourselves to handling user stuff.
 */
static int
initdev(DevicePtr X11devptr, usbsmPriv *privdata,char *xdevname)
{
	DeviceIntPtr devintp=(DeviceIntPtr)X11devptr;
	Atom tablet_atom;
	/* a "map" of button codes (HUH?) */
	CARD8 buttonmap[6]={0,1,2,3,4,5}; 
	int num_axises, num_buttons=5;
	int max_x, max_y;
	Bool am_focusable=FALSE;
	

	num_axises=2;
	max_x=10000; /* ignored, I think */
	max_y=10000; /* ignored, I think */


	/*******************************************************************
	 * Assume done with all the device driver opening, and             *
	 * internal status clearing. Now we can register with the X server *
	 *******************************************************************/

	tablet_atom=MakeAtom(xdevname,strlen(xdevname),FALSE);
	AssignTypeAndName(devintp, tablet_atom, xdevname);

	if(RegisterFdIo(X11devptr, privdata->fd,
                usbsmGetEvents, usbsmEnqueue) != Success)
	{
		ErrorF("BOLTusbsm Failed on RegisterFdIo\n");
		return(!Success);
	}

	/* This one call supposedly superceeds
		InitButtonClassDeviceStruct(),
		InitValuatorClassDeviceStruct(),
		InitPtrFeedbackclassDeviceStruct()
	 */

	if(InitPointerDeviceStruct(X11devptr, buttonmap, num_buttons,
	                        usbsmMotionProc,
				usbsmPtrCtrlProc,
				0 /* numMotionEvents?? */
				) != TRUE )
	{
		ErrorF("InitPointerDeviceStruct() failed\n");
		return (!Success);
	}

	/* Things seem to actually work without this call just fine.
	 * But since the X Server Developer guide mentions we
	 * should do these calls, we do it.
	 */
	InitValuatorAxisStruct(devintp,0,0,max_x,max_x);
	InitValuatorAxisStruct(devintp,1,0,max_y,max_y);

	/* This says, "We can become the core pointer" */
	/* Although again, it seems not neccessary */
	RegisterXPointerInterest(X11devptr, am_focusable, usbsmAxisChange);

	return (Success);
	
}


int staticcount=0; /* This isnt really important any more */



/* BOLTusbsmProc - bolthole.com top-level entry point for this
 *  usbsm tablet ddx extension to Xsun.
 *
 * This can be considered what the Sun ddx docs refer to as
 *  "DeviceControlProc()"
 *
 * This routine gets called by the X server, by virtue of adding the
 * name of this function to OWConfig.
 *
 * Analogous to InitAndStartDevices(), in X11 XINPUT documentation. Sigh.
 *  The Xserver first calls us with devptr=??, mode=DEVICE_INIT.
 *
 */
int ddxBOLTusbsmProc(DevicePtr X11devptr, int mode)
{
	usbsmPriv *privdata;
	int numdevs=1;
	char *numdevstring;
	char *devstring="/dev/usb/usbsm0";
	char devname[20];
	int devid;

	/* This is more an example of how to get an attribute from OWconfig.
	 * It doesnt do anything useful any more
	 */
	numdevstring=OWconfigGetAttribute("XINPUT","usbsm","numdevs");
	if(numdevstring!=NULL){
		numdevs=atoi(numdevstring);
		if(numdevs<1) numdevs=1;
		OWconfigFreeAttribute(numdevstring);
	}

	if(X11devptr==NULL){
		ErrorF("ddxBOLTusbsmProc: passed NULL DevicePtr\n");
		return (!Success);
	}

	privdata=(usbsmPriv*)X11devptr->devicePrivate;
	

	switch(mode){

	    case DEVICE_INIT:

#ifdef DEBUG
		ErrorF("BOLTusbsmProc: %s\n",version);
		ErrorF("BOLTusbsmProc: OWconf says numdevs=%d\n",numdevs);
#else
		strlen(version); /* to make compiler be quiet about version */
#endif
		if(privdata != NULL)
		{
			/* only allowed to call init once */
			return (!Success);
		}

		/* We need to always set this on DEVICE_INIT.
		 * In preparation for a second call to this routine,
		 * with DEVICE_ON
		 */
		X11devptr->on=FALSE;

		/*****************************************
		 * This must be the first time we are called.
		 * set up for 'primary' id, which we assume
		 * will be for the 'pen'.
		 *
		 * Yes, this is a really, really creepy cast.
		 * But it somehow works. Based off the cast in
		 * the Enqueue() routine, which is actually based from
		 * Sun sample code.
		 */

		privdata=(usbsmPriv*)xalloc(sizeof(usbsmPriv));
		if(privdata==NULL) return (!Success);
		bzero(privdata,sizeof(usbsmPriv));


		devid=((DeviceIntPtr)X11devptr)->id;
#ifdef DEBUG
		ErrorF("Debug: BOLTusbsmProc called with devid= %d\n",
		       devid);
#endif

		X11devptr->devicePrivate=privdata;
		
		/* The devname should really be taken from the
		 * configs. arg.
		 */
		switch(opendev(X11devptr,devstring)){
			case 1:
				break;

			default:
				ErrorF("opendev returned unexpected val\n");
				xfree(privdata);
				return(!Success);
				/* Assume if we can open the device, we should
				* take the primary role, and if we cant,
				* then someone else already has it
				*/
		}

		sprintf(devname,"usbsmdev%d",staticcount++);

		
#ifdef DEBUG
		ErrorF("usbsm DEVICE_INIT calling initdev for %s\n",devname);
#endif
		initdev(X11devptr,privdata,devname);

		return (Success);

		break;

	    case DEVICE_ON:
		X11devptr->on = TRUE;


#ifdef DEBUG
		ErrorF("usbsm calling AddEnabledDevice for DEVICE_ON: fd=%d\n",
		       privdata->fd);
#endif
		AddEnabledDevice (privdata->fd);

		/* say "We want SIGPOLL" when new data arrives */

		ioctl(privdata->fd, I_SETSIG, S_INPUT);

#ifdef DEBUG
		ErrorF("usbsm done with ioctl I_SETSIG\n");
#endif

		
		break;
	    case DEVICE_OFF:
		ioctl(privdata->fd, I_SETSIG, NULL);

		X11devptr->on = FALSE;
		break;

	    case DEVICE_CLOSE:
		if(X11devptr->on) {
#ifdef DEBUG
			ErrorF("ddxBOLTusbsm  calling RemoveEnabledDevice\n");
#endif
			RemoveEnabledDevice (privdata->fd);
			close(privdata->fd);
		}
		xfree(X11devptr->devicePrivate);
		X11devptr->on = FALSE;
		break;
	}

	return (Success);
}


